@@ -75,6 +75,11 @@
 # define snprintf _snprintf
 #endif
 
+/* Define sig_atomic_t if it's not available in signal.h. */
+#ifndef HAVE_SIG_ATOMIC_T
+typedef int sig_atomic_t;
+#endif
+
 /* Windows does not define ssize_t. */
 #ifndef HAVE_SSIZE_T
 typedef ptrdiff_t ssize_t;
@@ -127,6 +132,9 @@ extern int vsnprintf(char *, size_t, const char *, va_list);
 #if !HAVE_DAEMON
 extern int daemon(int, int);
 #endif
+#if !HAVE_MKSTEMP
+extern int mkstemp(char *);
+#endif
 #if !HAVE_STRLCAT
 extern size_t strlcat(char *, const char *, size_t);
 #endif
