@@ -37,6 +37,8 @@
 #include <config.h>
 #include <portable/system.h>
 
+#include <assert.h>
+
 #include <util/vector.h>
 #include <util/xmalloc.h>
 
@@ -51,8 +53,8 @@ vector_new(void)
 
     vector = xmalloc(sizeof(struct vector));
     vector->count = 0;
-    vector->allocated = 0;
-    vector->strings = NULL;
+    vector->allocated = 1;
+    vector->strings = xcalloc(1, sizeof(char *));
     return vector;
 }
 
@@ -63,46 +65,43 @@ cvector_new(void)
 
     vector = xmalloc(sizeof(struct cvector));
     vector->count = 0;
-    vector->allocated = 0;
-    vector->strings = NULL;
+    vector->allocated = 1;
+    vector->strings = xcalloc(1, sizeof(const char *));
     return vector;
 }
 
 
 /*
- * Resize a vector (using realloc to resize the table).
+ * Resize a vector (using realloc to resize the table).  Maintain a minimum
+ * allocated size of 1 so that the strings data element is never NULL.  This
+ * simplifies other code.
  */
 void
 vector_resize(struct vector *vector, size_t size)
 {
     size_t i;
 
+    assert(vector != NULL);
     if (vector->count > size) {
         for (i = size; i < vector->count; i++)
             free(vector->strings[i]);
         vector->count = size;
     }
-    if (size == 0) {
-        free(vector->strings);
-        vector->strings = NULL;
-    } else {
-        vector->strings = xrealloc(vector->strings, size * sizeof(char *));
-    }
+    if (size == 0)
+        size = 1;
+    vector->strings = xrealloc(vector->strings, size * sizeof(char *));
     vector->allocated = size;
 }
 
 void
 cvector_resize(struct cvector *vector, size_t size)
 {
+    assert(vector != NULL);
     if (vector->count > size)
         vector->count = size;
-    if (size == 0) {
-        free(vector->strings);
-        vector->strings = NULL;
-    } else {
-        vector->strings =
-            xrealloc(vector->strings, size * sizeof(const char *));
-    }
+    if (size == 0)
+        size = 1;
+    vector->strings = xrealloc(vector->strings, size * sizeof(const char *));
     vector->allocated = size;
 }
 
@@ -117,6 +116,7 @@ vector_add(struct vector *vector, const char *string)
 {
     size_t next = vector->count;
 
+    assert(vector != NULL);
     if (vector->count == vector->allocated)
         vector_resize(vector, vector->allocated + 1);
     vector->strings[next] = xstrdup(string);
@@ -128,6 +128,7 @@ cvector_add(struct cvector *vector, const char *string)
 {
     size_t next = vector->count;
 
+    assert(vector != NULL);
     if (vector->count == vector->allocated)
         cvector_resize(vector, vector->allocated + 1);
     vector->strings[next] = string;
@@ -136,16 +137,17 @@ cvector_add(struct cvector *vector, const char *string)
 
 
 /*
- * Add a new counted string to the vector, resizing the vector as necessary
- * the same as with vector_add.  This function is only available for vectors,
- * not cvectors, since it requires the duplication of the input string to be
- * sure it's nul-terminated.
+ * Add a new string to the vector, copying at most length characters of the
+ * string, resizing the vector as necessary the same as with vector_add.  This
+ * function is only available for vectors, not cvectors, since it requires the
+ * duplication of the input string to be sure it's nul-terminated.
  */
 void
 vector_addn(struct vector *vector, const char *string, size_t length)
 {
     size_t next = vector->count;
 
+    assert(vector != NULL);
     if (vector->count == vector->allocated)
         vector_resize(vector, vector->allocated + 1);
     vector->strings[next] = xstrndup(string, length);
@@ -161,6 +163,7 @@ vector_clear(struct vector *vector)
 {
     size_t i;
 
+    assert(vector != NULL);
     for (i = 0; i < vector->count; i++)
         free(vector->strings[i]);
     vector->count = 0;
@@ -169,6 +172,7 @@ vector_clear(struct vector *vector)
 void
 cvector_clear(struct cvector *vector)
 {
+    assert(vector != NULL);
     vector->count = 0;
 }
 
@@ -179,6 +183,8 @@ cvector_clear(struct cvector *vector)
 void
 vector_free(struct vector *vector)
 {
+    if (vector == NULL)
+        return;
     vector_clear(vector);
     free(vector->strings);
     free(vector);
@@ -187,6 +193,8 @@ vector_free(struct vector *vector)
 void
 cvector_free(struct cvector *vector)
 {
+    if (vector == NULL)
+        return;
     cvector_clear(vector);
     free(vector->strings);
     free(vector);
@@ -194,8 +202,8 @@ cvector_free(struct cvector *vector)
 
 
 /*
- * Given a vector that we may be reusing, clear it out.  If the first argument
- * is NULL, allocate a new vector.  Used by vector_split*.
+ * Given a vector that we may be reusing, clear it out.  If the argument is
+ * NULL, allocate a new vector.  Helper function for vector_split*.
  */
 static struct vector *
 vector_reuse(struct vector *vector)
@@ -232,7 +240,7 @@ split_count(const char *string, char separator)
 
     if (*string == '\0')
         return 1;
-    for (count = 1, p = string; *p; p++)
+    for (count = 1, p = string; *p != '\0'; p++)
         if (*p == separator)
             count++;
     return count;
@@ -241,8 +249,9 @@ split_count(const char *string, char separator)
 
 /*
  * Given a string and a separator character, form a vector by splitting the
- * string at those separators.  Do a first pass to size the vector, and if the
- * third argument isn't NULL, reuse it.  Otherwise, allocate a new one.
+ * string at occurrences of that separator.  Consecutive occurrences of the
+ * character will result in empty strings added to the vector.  Reuse the
+ * provided vector if non-NULL.
  */
 struct vector *
 vector_split(const char *string, char separator, struct vector *vector)
@@ -250,20 +259,22 @@ vector_split(const char *string, char separator, struct vector *vector)
     const char *p, *start;
     size_t i, count;
 
+    /* If the vector argument isn't NULL, reuse it. */
     vector = vector_reuse(vector);
 
+    /* Do a first pass to size the vector. */
     count = split_count(string, separator);
     if (vector->allocated < count)
         vector_resize(vector, count);
 
-    for (start = string, p = string, i = 0; *p; p++)
+    /* Walk the string and create the new strings with xstrndup. */
+    for (start = string, p = string, i = 0; *p != '\0'; p++)
         if (*p == separator) {
             vector->strings[i++] = xstrndup(start, p - start);
             start = p + 1;
         }
     vector->strings[i++] = xstrndup(start, p - start);
     vector->count = i;
-
     return vector;
 }
 
@@ -271,9 +282,8 @@ vector_split(const char *string, char separator, struct vector *vector)
 /*
  * Given a modifiable string and a separator character, form a cvector by
  * modifying the string in-place to add nuls at the separators and then
- * building a vector of pointers into the string.  Do a first pass to size the
- * vector, and if the third argument isn't NULL, reuse it.  Otherwise,
- * allocate a new one.
+ * building a vector of pointers into the string.  Reuse the provided vector
+ * if non-NULL.
  */
 struct cvector *
 cvector_split(char *string, char separator, struct cvector *vector)
@@ -281,12 +291,18 @@ cvector_split(char *string, char separator, struct cvector *vector)
     char *p, *start;
     size_t i, count;
 
+    /* If the vector argument isn't NULL, reuse it. */
     vector = cvector_reuse(vector);
 
+    /* Do a first pass to size the vector. */
     count = split_count(string, separator);
     if (vector->allocated < count)
         cvector_resize(vector, count);
 
+    /*
+     * Walk the string and replace separators with nul, and store the pointers
+     * to the start of each string segment.
+     */
     for (start = string, p = string, i = 0; *p; p++)
         if (*p == separator) {
             *p = '\0';
@@ -295,7 +311,6 @@ cvector_split(char *string, char separator, struct cvector *vector)
         }
     vector->strings[i++] = start;
     vector->count = i;
-
     return vector;
 }
 
@@ -303,7 +318,8 @@ cvector_split(char *string, char separator, struct cvector *vector)
 /*
  * Given a string and a set of separators expressed as a string, count the
  * number of strings that it will split into when splitting on those
- * separators.
+ * separators.  Unlike with split_count, multiple consecutive separator
+ * characters will be treated the same as a single separator.
  */
 static size_t
 split_multi_count(const char *string, const char *seps)
@@ -311,8 +327,14 @@ split_multi_count(const char *string, const char *seps)
     const char *p;
     size_t count;
 
+    /* The empty string produces no substrings. */
     if (*string == '\0')
         return 0;
+
+    /*
+     * Walk the string looking for the first separator not preceeded by
+     * another separator (and ignore a separator at the start of the string).
+     */
     for (count = 1, p = string + 1; *p != '\0'; p++)
         if (strchr(seps, *p) != NULL && strchr(seps, p[-1]) == NULL)
             count++;
@@ -329,9 +351,8 @@ split_multi_count(const char *string, const char *seps)
 
 /*
  * Given a string, split it at any of the provided separators to form a
- * vector, copying each string segment.  If the third argument isn't NULL,
- * reuse that vector; otherwise, allocate a new one.  Any number of
- * consecutive separators are considered a single separator.
+ * vector, copying each string segment.  Any number of consecutive separators
+ * are considered a single separator.  Reuse the provided vector if non-NULL.
  */
 struct vector *
 vector_split_multi(const char *string, const char *seps,
@@ -340,12 +361,19 @@ vector_split_multi(const char *string, const char *seps,
     const char *p, *start;
     size_t i, count;
 
+    /* If the vector argument isn't NULL, reuse it. */
     vector = vector_reuse(vector);
 
+    /* Count the number of strings we'll create and size the vector. */
     count = split_multi_count(string, seps);
     if (vector->allocated < count)
         vector_resize(vector, count);
 
+    /*
+     * Walk the string and look for separators.  start tracks the
+     * non-separator that starts a new string, so as long as start == p, we're
+     * tracking a sequence of separators.
+     */
     for (start = string, p = string, i = 0; *p != '\0'; p++)
         if (strchr(seps, *p) != NULL) {
             if (start != p)
@@ -355,7 +383,6 @@ vector_split_multi(const char *string, const char *seps,
     if (start != p)
         vector->strings[i++] = xstrndup(start, p - start);
     vector->count = i;
-
     return vector;
 }
 
@@ -363,9 +390,8 @@ vector_split_multi(const char *string, const char *seps,
 /*
  * Given a string, split it at any of the provided separators to form a
  * vector, destructively modifying the string to nul-terminate each segment.
- * If the third argument isn't NULL, reuse that vector; otherwise, allocate a
- * new one.  Any number of consecutive separators are considered a single
- * separator.
+ * Any number of consecutive separators are considered a single separator.
+ * Reuse the provided vector if non-NULL.
  */
 struct cvector *
 cvector_split_multi(char *string, const char *seps, struct cvector *vector)
@@ -373,12 +399,20 @@ cvector_split_multi(char *string, const char *seps, struct cvector *vector)
     char *p, *start;
     size_t i, count;
 
+    /* If the vector argument isn't NULL, reuse it. */
     vector = cvector_reuse(vector);
 
+    /* Count the number of strings we'll create and size the vector. */
     count = split_multi_count(string, seps);
     if (vector->allocated < count)
         cvector_resize(vector, count);
 
+    /*
+     * Walk the string and look for separators, replacing the ones that
+     * terminate a substring with a nul.  start tracks the non-separator that
+     * starts a new string, so as long as start == p, we're tracking a
+     * sequence of separators.
+     */
     for (start = string, p = string, i = 0; *p != '\0'; p++)
         if (strchr(seps, *p) != NULL) {
             if (start != p) {
@@ -390,100 +424,34 @@ cvector_split_multi(char *string, const char *seps, struct cvector *vector)
     if (start != p)
         vector->strings[i++] = start;
     vector->count = i;
-
     return vector;
 }
 
 
-/*
- * Given a string, count the number of strings that it will split into when
- * splitting on whitespace.
- */
-static size_t
-split_space_count(const char *string)
-{
-    const char *p;
-    size_t count;
-
-    if (*string == '\0')
-        return 0;
-    for (count = 1, p = string + 1; *p != '\0'; p++)
-        if ((*p == ' ' || *p == '\t') && !(p[-1] == ' ' || p[-1] == '\t'))
-            count++;
-
-    /*
-     * If the string ends in whitespace, we've overestimated the number of
-     * strings by one.
-     */
-    if (p[-1] == ' ' || p[-1] == '\t')
-        count--;
-    return count;
-}
-
-
 /*
  * Given a string, split it at whitespace to form a vector, copying each
- * string segment.  If the second argument isn't NULL, reuse that vector;
- * otherwise, allocate a new one.  Any number of consecutive whitespace
- * characters are considered a single separator.
+ * string segment.  Any number of consecutive whitespace characters are
+ * considered a single separator.  Reuse the provided vector if non-NULL.
+ * This is just a special case of vector_split_multi.
  */
 struct vector *
 vector_split_space(const char *string, struct vector *vector)
 {
-    const char *p, *start;
-    size_t i, count;
-
-    vector = vector_reuse(vector);
-
-    count = split_space_count(string);
-    if (vector->allocated < count)
-        vector_resize(vector, count);
-
-    for (start = string, p = string, i = 0; *p != '\0'; p++)
-        if (*p == ' ' || *p == '\t') {
-            if (start != p)
-                vector->strings[i++] = xstrndup(start, p - start);
-            start = p + 1;
-        }
-    if (start != p)
-        vector->strings[i++] = xstrndup(start, p - start);
-    vector->count = i;
-
-    return vector;
+    return vector_split_multi(string, " \t", vector);
 }
 
 
 /*
  * Given a string, split it at whitespace to form a vector, destructively
- * modifying the string to nul-terminate each segment.  If the second argument
- * isn't NULL, reuse that vector; otherwise, allocate a new one.  Any number
- * of consecutive whitespace characters are considered a single separator.
+ * modifying the string to nul-terminate each segment.  Any number of
+ * consecutive whitespace characters are considered a single separator.  Reuse
+ * the provided vector if non-NULL.  This is just a special case of
+ * cvector_split_multi.
  */
 struct cvector *
 cvector_split_space(char *string, struct cvector *vector)
 {
-    char *p, *start;
-    size_t i, count;
-
-    vector = cvector_reuse(vector);
-
-    count = split_space_count(string);
-    if (vector->allocated < count)
-        cvector_resize(vector, count);
-
-    for (start = string, p = string, i = 0; *p != '\0'; p++)
-        if (*p == ' ' || *p == '\t') {
-            if (start != p) {
-                *p = '\0';
-                vector->strings[i++] = start;
-            }
-            start = p + 1;
-        }
-    if (start != p)
-        vector->strings[i++] = start;
-    vector->count = i;
-
-    return vector;
+    return cvector_split_multi(string, " \t", vector);
 }
 
 
@@ -498,20 +466,24 @@ vector_join(const struct vector *vector, const char *seperator)
     char *string;
     size_t i, size, seplen;
 
+    /* If the vector is empty, this is trivial. */
+    assert(vector != NULL);
     if (vector->count == 0)
         return xstrdup("");
+
+    /* Determine the total size of the resulting string. */
     seplen = strlen(seperator);
     for (size = 0, i = 0; i < vector->count; i++)
         size += strlen(vector->strings[i]);
     size += (vector->count - 1) * seplen + 1;
 
+    /* Allocate the memory and build up the string using strlcat. */
     string = xmalloc(size);
     strlcpy(string, vector->strings[0], size);
     for (i = 1; i < vector->count; i++) {
         strlcat(string, seperator, size);
         strlcat(string, vector->strings[i], size);
     }
-
     return string;
 }
 
@@ -521,32 +493,38 @@ cvector_join(const struct cvector *vector, const char *seperator)
     char *string;
     size_t i, size, seplen;
 
+    /* If the vector is empty, this is trivial. */
+    assert(vector != NULL);
     if (vector->count == 0)
         return xstrdup("");
+
+    /* Determine the total size of the resulting string. */
     seplen = strlen(seperator);
     for (size = 0, i = 0; i < vector->count; i++)
         size += strlen(vector->strings[i]);
     size += (vector->count - 1) * seplen + 1;
 
+    /* Allocate the memory and build up the string using strlcat. */
     string = xmalloc(size);
     strlcpy(string, vector->strings[0], size);
     for (i = 1; i < vector->count; i++) {
         strlcat(string, seperator, size);
         strlcat(string, vector->strings[i], size);
     }
-
     return string;
 }
 
 
 /*
  * Given a vector and a path to a program, exec that program with the vector
- * as its arguments.  This requires adding a NULL terminator to the vector and
- * casting it appropriately.
+ * as its arguments.  This requires adding a NULL terminator to the vector
+ * (which we do not add to count, so it will be invisible to other users of
+ * the vector) and casting it appropriately.
  */
 int
 vector_exec(const char *path, struct vector *vector)
 {
+    assert(vector != NULL);
     if (vector->allocated == vector->count)
         vector_resize(vector, vector->count + 1);
     vector->strings[vector->count] = NULL;
@@ -556,6 +534,7 @@ vector_exec(const char *path, struct vector *vector)
 int
 cvector_exec(const char *path, struct cvector *vector)
 {
+    assert(vector != NULL);
     if (vector->allocated == vector->count)
         cvector_resize(vector, vector->count + 1);
     vector->strings[vector->count] = NULL;
